# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

bl_info = {
    "name": "Handle Panel",
    "description": "Displays control panels for 'handle objects', found in current scene",
    "author": "Witold Jaworski",
    "version": (1, 0, 2),
    "blender": (2, 80, 3),
    "location": "Properties > Scene ",
    "support": "COMMUNITY",
    "category": "Scene",
    "warning": "",
    "wiki_url": "http://airplanes3d.net/tutorials-020_e.xml",
    "tracker_url": "http://airplanes3d.net/track-020_e.xml"
    }
#----- updates
#2019-08-23: adaptation for Blender 2.8
#----- imports
import bpy
from bpy.utils import unregister_class, register_class
from bpy.props import FloatProperty, StringProperty
from mathutils import Vector


#----- diagnostic flag 
DEBUG = 0 #set it to 0 in the productive version, > 0 to see diagnosic messages, 2 to enabe PyDev debugger
###--- for direct debugging of this add-on (update the pydevd path!) ---------------------------
if DEBUG >= 2:
    import sys
    pydev_path = 'C:/Users/me/.p2/pool/plugins/org.python.pydev.core_7.2.1.201904261721/pysrc'
    if sys.path.count(pydev_path) < 1: sys.path.append(pydev_path)
    import pydevd
    pydevd.settrace(stdoutToServer=True, stderrToServer=True, suspend=False) #stop at first breakpoint
    #Beware: remove all breakpoints from other PyDev projects opened in Eclipse IDE
    
#----- constants 
PROP_LABEL  = "HANDLE:label"            #Label of the handle, to be displayed
PANEL_ID    = "SCENE_PT_HandlePanel%d"  #bl_idname format for subsequent panel classes
HANDLE_SUFFIX = ".Handle"               #we are looking for object having such names
#----- utility functions
def is_handle_name(name):
    ''' Returns True, when name is named as a handle
        Arguments:
        @name (string): an object name
    '''
    result = name[-len(HANDLE_SUFFIX):] == HANDLE_SUFFIX # *.Handle
    if not result:
        result = name[-len(HANDLE_SUFFIX)-4:-4] == HANDLE_SUFFIX and name[-4:] >= '.000' and name[-4:] <= '.999' #*.Handle.000 ... *.Handle.999
    return result

def get_frame(handle):
    ''' Returns (min, max) tuple for the handle object
        Arguments:
        @handle (bpy.types.Object): object - assumed that has valid limit constraint
    '''
    constr = SceneRefreshHandles.get_limit(handle)
    min = Vector((constr.min_x, constr.min_y, constr.min_z))
    max = Vector((constr.max_x, constr.max_y, constr.max_z))
    return (min, max)

def get_pos(handle, loc):
    ''' Returns position of specified point of the handle - a float within [0.. 1] range
        Arguments:
        @handle (bpy.types.Object): object - assumed that has valid limit constraint
        @loc (mathutils.Vector): a 3D vector, assumed to lie somewhere between handle limits 
    '''
    min, max = get_frame(handle)
    l = (max - min).length
    v = (loc - min).length
    if l == 0.0: 
        return 0
    else:
        return 100.0*v/l

def set_pos(handle, v):
    ''' Sets position of the handle to point determined by share v
        Arguments:
        @handle (bpy.types.Object): object - assumed that has valid limit constraint
        @v (float): a value within [0... 100] range 
    '''
    v /= 100.0
    min, max = get_frame(handle)
    loc = min*(1.0 - v) + max*(v)
    handle.location = loc

#----- public classes
class HandlePanel: #base class for all handle panels
    #bl_idname = PANEL_ID    - set in in the descendant classes
    bl_label = " "
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "scene"
    
    #--- helper properties
    object_id = "" #The id of the assigned handle object (string). Set it in descendant classes
    
    #--- helper functions
    @classmethod
    def get_label(cls, objects):
        '''Returns the object label - creates it on the object, if needed
            Arguments:
            @objects(collection): collection of Blender objects
        '''
        object = objects[cls.object_id]
        if PROP_LABEL not in object: #default label: the name of the object
            object[PROP_LABEL] = object.name
        return object[PROP_LABEL]
    
    #--- Blender interface methods    
    @classmethod
    def poll(cls, context):
        return (context.mode == 'OBJECT' and cls.object_id in bpy.context.scene.objects)
    
    def draw_header(self, context):
        row = self.layout.row()
        row.label(text = " " + self.get_label(context.scene.objects) + "       ", icon='SETTINGS')
        

    def draw(self, context):
        object = context.scene.objects[self.object_id]
        layout = self.layout.row()
        row = layout.split(factor = 1.0)
        row.prop(object, "handle_panel_pos", slider=True)
        row = layout.split(factor = 0.5) #when you have three buttons -set it to 0.33
        prop = row.operator(HandleReset.bl_idname, text="", icon='PAUSE')
        prop.handle_id = self.object_id
        prop = row.operator(HandleLabel.bl_idname, text="", icon='OUTLINER_DATA_GP_LAYER')
        prop.handle_id = self.object_id
        

class HandleControlPanel(bpy.types.Panel):
    bl_idname = "SCENE_PT_HandleControlPanel"
    bl_label = "Handle Panel"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "scene"

    #--- methods    
    @classmethod
    def poll(cls, context):
        #show this panel in Object Mode, only:
        return (context.mode == 'OBJECT')
        
    def draw(self, context):
        layout = self.layout
        layout = layout.row()
        layout.operator(SceneRefreshHandles.bl_idname, icon='FILE_REFRESH')
        if SceneRefreshHandles.panels:
            descr = " {0:d} handles found:".format(len(SceneRefreshHandles.panels))
            layout.label(text = descr, icon = 'INFO')
        else:
           descr = " Press this button to start..."
           layout.label(text = descr, icon = 'BACK')
     

class SceneRefreshHandles(bpy.types.Operator):
    '''Creates a control panel for each handle object, found in current scene'''
    bl_idname = "scene.refresh_handles"
    bl_label = "Show/Refresh"
    bl_description = "Create or refresh the collection of handle panels (displayed below)"
    
    #--- helper data
    panels = [] #list of registered panel classes
    
    #--- helper functions
    @staticmethod
    def get_limit(object):
        '''Returns the LimitLocation, assigned to <object>, or None
            Arguments:
            @object (bpy.types.Object): object to be searched
        '''
        if object.parent: #we are looking for objects with parents,only:
            limits = filter(lambda c: c.type == 'LIMIT_LOCATION', object.constraints)
            return next(limits,None) #assume that our constraint is the first limit
        else:
            return None
            
    @classmethod
    def unregister_panels(cls):
        for panel in cls.panels:
            unregister_class(panel)

    @classmethod
    def register_panels(cls):
        for panel in cls.panels:
            register_class(panel)
        
    @classmethod
    def create_panels(cls, objects):
        '''Returns list of newly created panel classes (may be empty)
            Arguments:
            @objects (collection): all scene objects
        '''
        cls.panels = []
        handles = [object for object in \
                    filter(lambda o: is_handle_name(o.name) and cls.get_limit(o), objects)]
        
        for object in handles:
            #synchronize the value to be displayed with actual handle location:
            object.handle_panel_pos = get_pos(object, object.location) #this is our add-on property!
            #create a dynamic panel class for this object:
            index = len(cls.panels)+1
            panel = type("HandlePanel%d" % index, (HandlePanel, bpy.types.Panel, ), \
                         {"object_id" : object.name, "bl_idname" : PANEL_ID % index}  \
                      )
            #store this type - to register/unregister
            cls.panels.append(panel)
            
            
        cls.panels.sort(key=lambda panel: panel.get_label(objects))
        
    #--- Blender interface methods        
    def execute(self, context):
        self.unregister_panels()
        
        self.create_panels(context.scene.objects)
        
        self.register_panels()
        
        context.area.tag_redraw()
        
        return {'FINISHED'}

class HandleReset(bpy.types.Operator):
    '''Resets position of the handle'''
    bl_idname = "scene.handle_reset"
    bl_label = "Reset"
    bl_description = "Reset handle to the neutral position"
    
    #--- parameters
    handle_id : StringProperty(name="Handle", description="Name of the handle object to be updated")

    #--- Blender interface methods
    def execute(self, context):
        if self.handle_id in context.scene.objects:
            handle = context.scene.objects[self.handle_id]
            loc = handle.location.copy()
            mat = handle.matrix_local.copy() #transformation "from parent to object"
            handle.handle_panel_pos = get_pos(handle, loc - mat.to_translation())
            
        return {'FINISHED'}

class HandleLabel(bpy.types.Operator):
    '''Opens a dialog box to update the label'''
    bl_idname = "scene.handle_label"
    bl_label = "Object details:"
    bl_description = "Change the label of this handle"
    
    #--- parameters
    handle_id : StringProperty(name="Object", description="Name of the handle object, to be updated")
    label : StringProperty(name="Label", description="Change description of this handle panel")
    
    #--- Blender interface methods
    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row = row.split(factor = 0.25)        
        row.label(text = "Label:")
        row = row.split(factor = 1.0)        
        row.prop(self, "label", text="")
        row = layout.row()
        row = row.split(factor = 0.25)        
        row.label(text = "ID name:")
        row = row.split(factor = 1.0)        
        row.label(text = self.handle_id)
        
    def invoke(self, context, event):
        if self.handle_id in context.scene.objects:
            object = context.scene.objects[self.handle_id]
            self.label = object[PROP_LABEL]
            return context.window_manager.invoke_props_dialog(self, width=350)
        else:
            return {'FINISHED'}
            
    def execute(self, context):
        if self.handle_id in context.scene.objects:
            object = context.scene.objects[self.handle_id]
            object[PROP_LABEL] = self.label
            context.area.tag_redraw()
        return {'FINISHED'}

# Updates the position of handle object 
def on_update(self, context):
    """ Called when the custom property - Object.handle_panel_pos - is changed
        Updates location of the object, according the new value
        Arguments:
        @self(bpy.types.Object): the object that owns the property
        @context(bpy.types.Context): context of the call
    """
    set_pos(self, self.handle_panel_pos)

#--- ### Register
#list of the classes in this add-on to be registered in Blender API:
            
classes = [
            HandleReset,
            HandleLabel,       
            SceneRefreshHandles,
            HandleControlPanel, 
          ]  

def register():
    for cls in classes:
        register_class(cls)
    #first: check, if our property has been added to Object class:
    if "handle_panel_pos" not in dir(bpy.types.Object):
       bpy.types.Object.handle_panel_pos = FloatProperty(name="Pos", \
                                                         description="Position on Handle Panel within its limits (see addon)", \
                                                         min=0.0, max=100, subtype='PERCENTAGE', update=on_update)
    if DEBUG: print(__name__ + ": registered")

def unregister():
    if "handle_panel_pos" in dir(bpy.types.Object):
       del bpy.types.Object.handle_panel_pos  #clean-up: we do not need to store it! 
        
    SceneRefreshHandles.unregister_panels()
    SceneRefreshHandles.panels = []
    for cls in classes:
        unregister_class(cls)  
    if DEBUG: print(__name__ + ": UNregistered")

if __name__ == '__main__':
    register()
